module Joebot.Types

open System

type ActionClass =
    | Network
    | Exploit
    | Penetration
 
type IClass = abstract GetClass : unit -> ActionClass
 
type Weapon =
    | Virus
    | Ransom
    | Worm
    | DDos
    | Crack
    | Injection
    interface IClass with
        member this.GetClass () =
            match this with
            | Virus | Ransom -> Exploit
            | DDos | Worm -> Network
            | Crack | Injection -> Penetration
    static member TryParse weapon =
        match weapon with
        | "Virus"     -> Some Virus
        | "Ransom"    -> Some Ransom    
        | "Worm"      -> Some Worm
        | "DDos"      -> Some DDos
        | "Crack"     -> Some Crack
        | "Injection" -> Some Injection
        | _ -> None
            
    

type Shield =
    | Firewall
    | PortScan
    | Encryption
    | Cypher
    | Hardening
    | Sanitation
    interface IClass with
        member this.GetClass () =
            match this with
            | Firewall | PortScan -> Exploit
            | Encryption | Cypher -> Network
            | Hardening | Sanitation -> Penetration
    static member TryParse shield =
        match shield with
        | "Firewall"   -> Some Firewall
        | "PortScan"   -> Some PortScan   
        | "Encryption" -> Some Encryption
        | "Cypher"     -> Some Cypher
        | "Hardening"  -> Some Hardening
        | "Sanitation" -> Some Sanitation
        | _ -> None

type HackResult =
    | Strong
    | Weak


type DiscordPlayer = {
    Id : uint64
    Name : string
}

type Attack = {
    HackType : Weapon
    Target : DiscordPlayer
    Timestamp : DateTime
}
    
type Defense = {
    DefenseType : Shield
    Timestamp : DateTime
}
    
type Player = {
    DiscordId : uint64
    Weapons : Weapon list
    Shields : Shield list
    Attacks : Attack list
    Defenses : Defense list
    Bank : single
}
 

